#include <assert.h>
#include <stdio.h>
#include "cstd.h"
#include "ff2k.h"

/*
    Copyright 2011-2012 Andrew V. Sutherland
    See LICENSE file for license details.
*/


// primitive polys taken from Hansen & Mullen Supplement in Math Comp Vol 59, No 200, 1992, pp. S47-S50
// each poly is (1<<k)+ff2k_xkrtab[k]-- we require the coefficients of x^n to be zero for floor(k/2)<n<k.
static ff2k_t ff2k_xkrtab[FF2K_MAXK+1] =
{ 0, 0, 0x3, 0x3, 0x3, 0x5, 0x3, 0x3, 0x1d, 0x11, 0x9, 0x5, 0x53, 0x1b, 0x2b, 0x3, 0x2d,
  0x9, 0x81, 0x27, 0x9, 0x5, 0x3, 0x23, 0x1b,  0x9, 0x47, 0x27, 0x9, 0x5, 0x53, 0x5 };
ff2k_t _ff2k_xk, _ff2k_xkr, _ff2k_x2km2, _ff2k_x2km2r, _ff2k_m, _ff2k_tr, _ff2k_f;
int _ff2k_k;

void ff2k_setup (int k)
{
    register int i, j;
    register ff2k_t a,t;
    
	assert ( k >= 1 );
	assert ( k <= FF2K_MAXK );
	_ff2k_k = k;
	_ff2k_xk = 1UL << k;
	_ff2k_m = _ff2k_xk-1;
	_ff2k_xkr = ff2k_xkrtab[k];
    _ff2k_f = _ff2k_xk | _ff2k_xkr;
	_ff2k_x2km2 = 1UL << (2*k-2);
	_ff2k_x2km2r = _ff2k_xkr << (k-2);
    // compute vector of monomial traces
    _ff2k_tr = 0 ;
    for ( i = 0 ; i < k ; i++ ) {
        a = ((ff2k_t)1)<<i;
        for ( t = a, j = 1; j < k ; j++ ) { _ff2k_square(a,a); _ff2k_addto(t,a); }
        if ( ! t ) continue;
        assert ( t==1 );
        _ff2k_tr |= ((ff2k_t)1)<<i;
    }
}

void ff2k_poly_print (ff2k_t f[], int d)
{
	register int i;
	
	printf ("[");
	for ( i = d ; i >= 0 ; i-- ) {
		if ( f[i] ) { 
			if ( i < d ) printf (" + "); 
			switch (i) { case 0: ff2k_print(f[i]); break; case 1: ff2k_print(f[i]); printf ("*X"); break; default: ff2k_print (f[i]); printf ("*X^%d", i); }
		}
	}
	puts ("]");
}


// pure brute-force root-finding, only for use with very small k (e.g. <= 3)
int ff2k_poly_distinct_roots (ff2k_t r[], ff2k_t f[], int d)
{
	ff2k_t x, t;
	register int i;
	
	_ff2k_set_zero(x);
	i = 0;
	do {
		t = ff2k_poly_eval (f, d, x);
		if ( _ff2k_zero(t) ) { _ff2k_set(r[i],x); i++; }
		_ff2k_next(x);
	} while ( ! _ff2k_zero(x) );
	return i;
}

// determine whether the elliptic curve with Weierstrass coefficients w = [a1,a2,a3,a4,a6] is non-singular
int ff2k_WS_nonsingular (ff2k_t w[5])
{
	register ff2k_t a1, a3, t1, t2, t3;
	
	if ( _ff2k_zero(w[0]) ) return ( _ff2k_zero(w[2]) ? 0 : 1);
	
	// compute D = a1^6*a6 + a1^5*a3*a4 + a1^4*a2*a3^2 + a1^4*a4^2 + a1^3*a3^3 + a3^4
	//                   = a1*(a1*(a1*(a1*(a1*(a1*a6+a3*a4)+a2*a3^2+a4^2)+a3^3))) + a3^4
	_ff2k_set (a1,w[0]); _ff2k_set (a3,w[2]);
	_ff2k_mult (t1,a1,w[4]); _ff2k_mult (t2,a3,w[3]); _ff2k_addto (t1,t2);
	_ff2k_mult (t1,t1,a1); _ff2k_square (t3,a3); _ff2k_mult(t2,w[1],t3); _ff2k_addto (t1,t2); _ff2k_square (t2, w[3]); _ff2k_addto (t1, t2);
	_ff2k_mult (t1,t1,a1); _ff2k_mult (t3,t3,a3); _ff2k_addto (t1,t3);
	_ff2k_square (t2, a1); _ff2k_mult (t2,t2,a1); _ff2k_mult (t1, t1, t2); _ff2k_mult(t3,t3,a3); _ff2k_addto (t1,t3);
	return ( _ff2k_zero (t1) ? 0 : 1 );
}

// very naive point-counting for elliptic curves over F_2^k, intended for small k (e.g. 1 or 2)
long ff2k_WS_pointcount (ff2k_t w[])
{
	register ff2k_t x, y, t1, t2;
	long pts;

	if ( ! ff2k_WS_nonsingular (w) ) return -1;
	pts = 0;
	_ff2k_set_zero(y);
	do {
		_ff2k_set_zero (x);
		do {
			_ff2k_square (t1, y);  _ff2k_mult (t2, w[0], y); _ff2k_mult (t2, t2, x); _ff2k_addto (t1, t2); _ff2k_mult (t2, w[2], y); _ff2k_addto (t1, t2);
			_ff2k_add (t2, x, w[1]); _ff2k_mult (t2, t2, x); _ff2k_addto (t2, w[3]); _ff2k_mult (t2, t2, x); _ff2k_addto (t2, w[4]);
			if ( _ff2k_equal (t1, t2) ) pts++;
			_ff2k_next(x);
		} while ( ! _ff2k_zero (x) );
		_ff2k_next(y);
	} while ( ! _ff2k_zero(y) );
//printf ("pointcount over F_%ld is %ld\n", (1L<<_ff2k_k), pts+1);
	return pts+1;	// always one point at infinity
}

// counts points on y^2+h(x)*y=f(x) over current finite field with 2^k elements
// does not check for good reduction
long ff2k_hyperelliptic_pointcount (ff2k_t f[], int df, ff2k_t h[], int dh)
{
	ff2k_t t1, t2, t3, x, y;
	long pts;
	
	pts = 0;
	_ff2k_set_zero (y);
	do {
		_ff2k_set_zero (x);
		do {
			t1=ff2k_poly_eval (f, df, x);
			t2=ff2k_poly_eval (h, dh, x);
			_ff2k_mult (t3, t2, y);
			_ff2k_square (t2, y);
			_ff2k_addto (t2, t3);
			if ( _ff2k_equal (t1, t2) ) pts++;
			_ff2k_next (x);
		} while ( ! _ff2k_zero (x) );
		_ff2k_next(y);
	} while ( ! _ff2k_zero (y) );
	return pts+ff2k_hyperelliptic_pointcount_at_infinity (df, dh, _ff2k_k);
}

// The root counting code below is based on Zinoviev's "On the solutions of equations of degree" RR-2829, 1996, inria-00073862
// https://hal.inria.fr/inria-00073862

int ff2k_poly_count_distinct_roots_d2 (ff2k_t f[3])
{
    register ff2k_t a, b, c;
    if ( ! _ff2k_one(f[2]) ) {
        _ff2k_invert(c,f[2]);  _ff2k_mult(a,f[1],c);  _ff2k_mult(b,f[0],c);
    } else {
        _ff2k_set(a,f[1]); _ff2k_set(b,f[0]);
    }
    if ( _ff2k_zero(a) ) return 1;
    _ff2k_square (c, a);
    _ff2k_mult(c,b,ff2k_invert(c));
    return _ff2k_zero(ff2k_trace(c)) ? 2 : 0;
}


// transpose the matrix -- this is seriously painful, but so is column reducing :(
// probably could be sped up by 2x (see Hacker's Delight pp 108-112, but not clear this is a bottleneck
void ff2_transpose (unsigned T[], unsigned M[], int rows, int cols)
{
    register unsigned m, x;
    register int i, j;
    
    for ( i = 0 ; i < cols ; i++ ) {
        m = ((unsigned)1)<<i;
        for ( x = 0, j = 0 ; j < rows ; j++ ) x |= ((M[j]&m)>>i)<<j;
        T[i] = x;
    }
}

// row reduce a matrix over GF(2) whose rows are bitstrings stored in ints
// reteurns the rank of the matrix
int ff2_row_reduce (unsigned M[], int rows, int cols)
{
    register unsigned m, x;
    register int r, j, k;

    assert (cols < 32);
    r = -1;
    for ( m = (unsigned)1 ; m <(1U<<cols) ; m<<= 1 ) {
        for ( j = r+1 ; j < rows ; j++ ) {
            if ( (M[j]&m) ) {
                r++;
                x = M[r]; M[r] = M[j]; M[j] = x;
                for ( k = r+1 ; k < rows ; k++ ) if ( (M[k]&m) ) M[k] ^= M[r];
            }
        }
    }
    return r+1;
}

// row reduce a matrix over GF(2) whose rows are bitstrings stored in ints
// sets R to a matrix representing row ops performed (so multiplying original M on the left by R yields reduced M)
// reteurns the rank of the matrix
int ff2_row_reduce_mat (unsigned R[], unsigned M[], int rows, int cols)
{
    register unsigned m, x;
    register int r, j, k;

    assert (cols < 32);
    for ( j = 0 ; j < rows ; j++ ) R[j] = (1U<<j);
    r = -1;
    for ( m = (unsigned)1 ; m <(1U<<cols) ; m<<= 1 ) {
        for ( j = r+1 ; j < rows ; j++ ) {
            if ( (M[j]&m) ) {
                r++;
                x = M[r]; M[r] = M[j]; M[j] = x;
                x = R[r]; R[r] = R[j]; R[j] = x;
                for ( k = r+1 ; k < rows ; k++ ) if ( (M[k]&m) ) { M[k] ^= M[r]; R[k] ^= R[r]; };
            }
        }
    }
    return r+1;
}

// Given linear poly L(z)=z^4+a*z^2+b*z return k*k matrix over GF(2) with rows L(z^i)
void ff2k_Lmatrix_slow (unsigned M[], ff2k_t a, ff2k_t b)
{
    register ff2k_t t, x, x2, x4, z, z2, z4;
    register int i;
    
    assert ( _ff2k_k <= 32 );
    _ff2k_add(t,a,b);  _ff2k_inc(t);   // L(z^0)
    M[0] = t;
    if ( _ff2k_k == 1 ) return;
    x = ((ff2k_t)1)<<1;  // z^1;
    _ff2k_square(x2,x);  _ff2k_square(x4,x2);
    _ff2k_mult(z,x,b);  _ff2k_mult(z2,x2,a);  _ff2k_set(z4,x4);
    _ff2k_add(t,z,z2);  _ff2k_addto(t,z4); 
    M[1] = t;
    for ( i = 2 ; i < _ff2k_k ; i++ ) {
        _ff2k_multby(z,x);  _ff2k_multby(z2,x2);  _ff2k_multby(z4,x4);  // these are really just shifts
        _ff2k_add(t,z,z2);  _ff2k_addto(t,z4);  M[i] = t;
    }
    return;
}

// Given linear poly L(z)=z^4+a*z^2+b*z return k*k matrix over GF(2) with rows L(z^i)
void ff2k_Lmatrix_fast (unsigned M[], ff2k_t a, ff2k_t b)
{
    register ff2k_t z, z2, z4;
    register int i;
    
    assert ( _ff2k_k > 4 && _ff2k_k <= 32 );
    M[0] = (a^b)^1;  // L(z^0)
    z = _ff2k_reduce(b<<1);  z2 = _ff2k_reduce(a<<2);  z4 = _ff2k_reduce(((ff2k_t)1)<<4);
    M[1] = (z^z2)^z4;
    for ( i = 2 ; i < _ff2k_k ; i++ ) {
        z = _ff2k_reduce(z<<1);  z2 = _ff2k_reduce(z2<<2);  z4 = _ff2k_reduce(z4<<4);
        M[i] = (z^z2)^z4;
    }
    return;
}
    
// binary materix-vec mult (n*n matrix times n-bit vector)
static inline unsigned ff2_mat_vec_mult (unsigned M[], int n, unsigned v)
    { register int i; register unsigned x; for ( i = 0, x = 0 ; i < n ; i++ ) x |= __builtin_parity(M[i]&(unsigned)v) << i; return x; }

// Attempts to tranform cubic f(x) into linearized L(x)=x^4+b*x^2+c*x+d with one more distinct roots
// If f(x) has a repeated root, the # of distinct roots is returned and L is not set, otherwise return value is 0 and L is set
int ff2k_poly_count_distinct_roots_d3_linearize (ff2k_t L[5], ff2k_t f[4])
{
    register ff2k_t a, b;

    if ( _ff2k_one(f[3]) ) {
        a = ff2k_add(ff2k_square(f[2]),f[1]);
        b = ff2k_add(ff2k_multiply(f[1],f[2]),f[0]);
    } else {
        a = ff2k_add(ff2k_square(f[2]),ff2k_multiply(f[1],f[3]));
        b = ff2k_multiply(f[3],ff2k_add(ff2k_multiply(f[1],f[2]),ff2k_multiply(f[0],f[3])));
    }

    // deal with repeated root cases
    if ( _ff2k_zero(b) ) return _ff2k_zero(a) ? 1 : 2;
    // now we have linearized poly x^4+a*x^2+b*x+? with no repeated roots and one extra root -- note that the constant term is irrelevant, so we set it to zero
    _ff2k_set_one(L[4]);  _ff2k_set_zero(L[3]);  _ff2k_set(L[2], a);  _ff2k_set (L[1],b); _ff2k_set_zero(L[0]);
    return 0;
}

int ff2k_poly_count_distinct_roots_d3 (ff2k_t f[4])
{
    unsigned M[FF2K_MAXK];
    ff2k_t L[5];
    register int r;

    r = ff2k_poly_count_distinct_roots_d3_linearize (L, f);
    if ( r ) return r;
    ff2k_Lmatrix(M,L[2],L[1]);
    r = ff2_row_reduce(M,_ff2k_k,_ff2k_k);
    return (1<<(_ff2k_k-r))-1;
}



// Attempts to tranform quartic f(x) into linearized L(x)=x^4+b*x^2+c*x+d with the same # distinct roots
// If f(x) has a repeated root, the # of distinct roots is returned and L is not set, otherwise return value is 0 and L is set
int ff2k_poly_count_distinct_roots_d4_linearize (ff2k_t L[5], ff2k_t f[5])
{
    register ff2k_t a, b, c, d, e, t;

    if ( ! _ff2k_one(f[4]) ) {
        _ff2k_invert(t,f[4]);  _ff2k_mult(a,f[3],t);  _ff2k_mult(b,f[2],t);  _ff2k_mult(c,f[1],t); _ff2k_mult(d,f[0],t);
    } else {
        _ff2k_set(a,f[3]); _ff2k_set(b,f[2]);  _ff2k_set(c,f[1]);  _ff2k_set(d,f[0]);
    }
    if ( ! _ff2k_zero(a) ) {
        // check discriminant a*(a*d+b*c)+c*c
        if ( _ff2k_zero(ff2k_add(ff2k_multiply(a,ff2k_add (ff2k_multiply(a,d),ff2k_multiply(b,c))),ff2k_square(c))) ) {
            if ( _ff2k_zero(ff2k_add(ff2k_multiply(a,c),ff2k_square(b))) ) return 2;
            _ff2k_set_one(L[2]);  _ff2k_set(L[1],a); _ff2k_invert(t,a); _ff2k_set(L[0],ff2k_add(b,ff2k_multiply(c,ff2k_invert(a))));
            return ff2k_poly_count_distinct_roots_d2(L)+1;
        }
        // apply substitution x -> x+e, where e = sqrt(c/a) to f(x) to kill linear term
        t = ff2k_multiply(c,ff2k_invert(a));    // e^2
        e = ff2k_sqrt(t);
        d = ff2k_add(d,ff2k_multiply(t,ff2k_add(b,t)));
        b = ff2k_add(b,ff2k_multiply(a,e));
        // now we have x^4+a*x^3+b*x^2+d, reverse polynomial and make it monic to get desired form
        d = ff2k_invert(d); // we could combine this with the inverse above, but it doesn' speed things up
        b = ff2k_multiply(b,d);
        c = ff2k_multiply(a,d);
    }
    _ff2k_set_one(L[4]);  _ff2k_set_zero(L[3]); _ff2k_set(L[2],b); _ff2k_set(L[1],c); _ff2k_set(L[0],d);
    return 0;
}

int ff2k_poly_count_distinct_roots_d4 (ff2k_t f[5])
{
    unsigned M[FF2K_MAXK];
    ff2k_t L[5];
    register int r;
    
    r = ff2k_poly_count_distinct_roots_d4_linearize (L, f);
    if ( r ) return r;
    r = ff2k_Rmatrix(M,L[2],L[1]);
    // Apply row-reduction ops to d to see if it is in the image of L(z) -- this will be true if  and only if the result is 0 past the rank.
    if ( (ff2_mat_vec_mult (M, _ff2k_k, L[0]) & ~((1u<<r)-1)) ) return 0;
    return (1<<(_ff2k_k-r));
}



