#ifndef _FF2K_INCLUDE_
#define _FF2K_INCLUDE_

/*
    Copyright 2011-2012 Andrew V. Sutherland
    See LICENSE file for license details.
*/

#include <gmp.h>
#include <stdint.h>
#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif

// very rudimentary implementation of arithmetic in F_{2^k}
// this is not meant to be fast, just functional

#define FF2K_MAXK			31		// it would not be hard to increase this to 63

typedef unsigned long ff2k_t;

extern int _ff2k_k;
extern ff2k_t _ff2k_xk, _ff2k_xkr, _ff2k_x2km2, _ff2k_x2km2r, _ff2km, _ff2k_m, _ff2k_tr, _ff2k_f;

void ff2k_setup (int k);

#define _ff2k_set(o,a)		((o) = (a))
#define _ff2k_equal(a,b)	((a) == (b))
#define _ff2k_add(o,a,b)	((o) = (a)^(b))
#define _ff2k_addto(o,a)	((o) ^= (a))
#define _ff2k_sub(o,a,b)	_ff2k_add(o,a,b)
#define _ff2k_subfrom(o,a)	_ff2k_addto(o,a)
#define _ff2k_inc(o)		((o) ^= 1UL)
#define _ff2k_zero(a)		(!(a))
#define _ff2k_set_zero(o)	((o)=0)
#define _ff2k_one(a)		((a) == 1UL)
#define _ff2k_set_one(o)	((o)=1UL)
#define _ff2k_negate(o)
#define _ff2k_next(o)		(++(o) == _ff2k_xk ? (o)=0 : (o))

#define _ff2k_pclmultby(a,b) asm ("pclmullqlqdq %1, %0;" : "+x"(a) : "x"(b));   // use Intel carry-less multiplication to compute a = a*b
static inline ff2k_t _ff2k_reduce(ff2k_t a)
{
    register ff2k_t q, r;

    // use Barret reduction without precomp (our primitive polys all have the upper half coeffs clear)
    // see https://www.cosic.esat.kuleuven.be/publications/article-1115.pdf
    q = a >> _ff2k_k; _ff2k_pclmultby(q,_ff2k_f);  q >>= _ff2k_k;
    r = a & _ff2k_m;  _ff2k_pclmultby(q,_ff2k_f);  q &= _ff2k_m; r ^= q;
    return r;
}

static inline ff2k_t ff2k_add (ff2k_t a, ff2k_t b) { return a^b; }
static inline ff2k_t ff2k_multiply (ff2k_t a, ff2k_t b)
{
    _ff2k_pclmultby(a,b);
    return _ff2k_reduce(a);
}
static inline ff2k_t ff2k_square (ff2k_t a) { return ff2k_multiply(a,a); }  // squaring is no faster than multiplying using carry-less mults

static inline ff2k_t ff2k_invert (ff2k_t a)
{
    register ff2k_t u, v, g1, g2;
    if ( _ff2k_zero(a) ) a=1/a; // force division by zero
    // this could be made faster but we don't use it much
    u = a; v = _ff2k_f;
    g1 = 1; g2 = 0;
    while ( u!=1 && v!=1 ) {
        while ( !(u&1) ) { u>>=1;  g1 = (g1&1) ? ((g1^_ff2k_f)>>1) : g1>>1; }
        while ( !(v&1) ) { v>>=1;  g2 = (g2&1) ? ((g2^_ff2k_f)>>1) : g2>>1; }
        if ( u > v ) { u ^=v; g1 ^= g2; } else { v ^= u; g2 ^= g1; }
    }
    return (u==1) ? g1 : g2;
}

static inline ff2k_t ff2k_trace (ff2k_t a)
    { return __builtin_parityl(a&_ff2k_tr); }

#define _ff2k_mult(o,a,b)		((o)=ff2k_multiply(a,b))
#define _ff2k_multby(o,a)		((o)=ff2k_multiply(o,a))
#define _ff2k_square(o,a)		((o)=ff2k_square(a))
#define _ff2k_invert(o,a)       ((o)=ff2k_invert(a))
#define _ff2k_trace(o,a)        ((o)=ff2k_trace(a))

// one can do this slightly more quickly, but this is fast enough
static inline ff2k_t ff2k_sqrt (ff2k_t a)
{
    register int i;
    ff2k_t x = a;
    for ( i = 0 ; i < _ff2k_k-1 ; i++ ) x = ff2k_square(x);
    return x;
}
    
static inline ff2k_t ff2k_primitive_root () { return (_ff2k_k == 1) ? 1UL : 2UL; }

static inline int ff2k_poly_set_mpz (ff2k_t f[], mpz_t F[], int d)
	{ register int i; for ( i = 0 ; i <= d ; i++ ) if ( mpz_tstbit(F[i],0) ) _ff2k_set_one(f[i]); else _ff2k_set_zero(f[i]);  for ( i-- ; i >= 0 && _ff2k_zero(f[i]) ; i-- ); return i; }

// computes o = f(x)
static inline ff2k_t ff2k_poly_eval (ff2k_t f[], int d, ff2k_t x)
{
	register ff2k_t y;
	register int i;
	
	if ( d < 0 ) { return 0; }
	_ff2k_set (y,f[d]);
	for ( i = d-1 ; i >= 0 ; i-- ) { _ff2k_multby(y, x); _ff2k_addto (y,f[i]); }
    return y;
}

// brute-force root-finding
int ff2k_poly_distinct_roots (ff2k_t r[], ff2k_t f[], int d);

// naive point-counting for elliptic curves in WS form over F_2^k, returns -1 if curve is singular
long ff2k_WS_pointcount (ff2k_t w[]);

// naive point-counting for hyperelliptic curves over F_2^k, assumes curve is not singular but does not check
long ff2k_hyperelliptic_pointcount (ff2k_t f[], int df, ff2k_t h[], int dh);

// returns # of points at infinity (assuming good reduction), 0, 1, or 2.
static inline int ff2k_hyperelliptic_pointcount_at_infinity (int df, int dh, int k)
	{ int g = ((df>2*dh?df:2*dh)-1)/2;  if ( dh <= g) return 1;  if ( df == 2*g+2 && (k&1) ) return 0;  return 2; }

static inline void ff2k_print (ff2k_t x)
	{ register int i; if ( _ff2k_zero(x) ) { printf ("(0)"); return; }printf ("("); for ( i = _ff2k_k-1 ; i >= 0 ; i-- ) { if ( x&(1<<i) ) switch (i) { case 0: printf("+1"); break; case 1: printf ("+z"); break; default: printf ("+z^%d", i); } } printf (")"); }
		
void ff2k_poly_print (ff2k_t f[], int d);

int ff2k_poly_count_distinct_roots_d2 (ff2k_t f[3]);
int ff2k_poly_count_distinct_roots_d3 (ff2k_t f[4]);
int ff2k_poly_count_distinct_roots_d4 (ff2k_t f[5]);

void ff2k_Lmatrix_slow (unsigned M[], ff2k_t a, ff2k_t b);    
void ff2k_Lmatrix_fast (unsigned M[], ff2k_t a, ff2k_t b);    
static inline void ff2k_Lmatrix (unsigned M[], ff2k_t a, ff2k_t b)
    { (_ff2k_k <= 4) ? ff2k_Lmatrix_slow(M,a,b) : ff2k_Lmatrix_fast(M,a,b); }

void ff2_transpose (unsigned T[], unsigned M[], int rows, int cols);
int ff2_row_reduce (unsigned M[], int rows, int cols);
int ff2_row_reduce_mat (unsigned R[], unsigned M[], int rows, int cols);
static inline int ff2k_Rmatrix (unsigned R[], ff2k_t a, ff2k_t b)
{
    unsigned M[FF2K_MAXK];
    ff2k_Lmatrix(R, a, b);
    ff2_transpose (M, R, _ff2k_k, _ff2k_k);
    return ff2_row_reduce_mat (R, M, _ff2k_k, _ff2k_k);
}

int ff2k_poly_count_distinct_roots_d3_linearize (ff2k_t L[5], ff2k_t f[4]);
int ff2k_poly_count_distinct_roots_d4_linearize (ff2k_t L[5], ff2k_t f[5]);


#ifdef __cplusplus
}
#endif

#endif
